import { useDocument, useDocumentOnce } from 'react-firebase-hooks/firestore';
import { useDocumentSnapData } from './useSnapData.js';
import { useFirestoreErrorLogger, useSafeRef } from './utils.js';
export const useTypedDocumentOnce = (typedDoc, { getOptions, ...dataOptions } = {}) => {
    const { safeRef, refChanged } = useSafeRef(typedDoc === null || typedDoc === void 0 ? void 0 : typedDoc.raw);
    const [_snap, _loading, error] = useDocumentOnce(safeRef, {
        getOptions,
    });
    const loading = _loading || refChanged;
    useFirestoreErrorLogger(error);
    const { snap, data } = useDocumentSnapData(typedDoc, _snap, dataOptions);
    return { snap, data, loading, error };
};
export const useTypedDocument = (typedDoc, { snapshotListenOptions, ...dataOptions } = {}) => {
    const { safeRef, refChanged } = useSafeRef(typedDoc === null || typedDoc === void 0 ? void 0 : typedDoc.raw);
    const [_snap, _loading, error] = useDocument(safeRef, {
        snapshotListenOptions,
    });
    const loading = _loading || refChanged;
    useFirestoreErrorLogger(error);
    const { snap, data } = useDocumentSnapData(typedDoc, _snap, dataOptions);
    return { snap, data, loading, error };
};
